package ch.othello.openscep;

import java.io.*;
import java.io.IOException;

import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.SecureRandom;

import org.bouncycastle.asn1.DERConstructedSequence;
import org.bouncycastle.asn1.DERInteger;
import org.bouncycastle.asn1.DERObject;
import org.bouncycastle.asn1.DERPrintableString;
import org.bouncycastle.asn1.DEROctetString;

import org.bouncycastle.asn1.pkcs.CertificationRequest;
import org.bouncycastle.asn1.pkcs.ContentInfo;
import org.bouncycastle.asn1.pkcs.RSAPrivateKeyStructure;
import org.bouncycastle.asn1.pkcs.SignedData;
import org.bouncycastle.asn1.util.ASN1Dump;
import org.bouncycastle.asn1.x509.*;

import ch.othello.openscep.internal.*;
import ch.othello.openscep.internal.commands.*;


/** Scep Client
 *
 * <pre>
 * $Id: ScepClient.java,v 1.54 2002/02/28 01:22:16 urut Exp $
 * $Log: ScepClient.java,v $
 * Revision 1.54  2002/02/28 01:22:16  urut
 * debug info
 *
 * Revision 1.53  2002/02/27 06:53:19  urut
 * modified setting of subject in Signed structure
 *
 * Revision 1.52  2002/02/27 04:21:27  urut
 * first working spkac version
 *
 * Revision 1.51  2002/02/27 04:11:09  alani
 * loadCA now only accepts a String
 *
 * Revision 1.50  2002/02/25 04:45:35  urut
 * set subjectScep from cmd line to test out various effects
 *
 * Revision 1.49  2002/02/25 04:23:39  urut
 * first working version for MS browsers
 * there is still a problem in here as subjectSCEP needs to be set to the subject
 * of the actual PKCS#10 request. seems to be a bug in OpenScep.
 *
 * Revision 1.48  2002/02/24 22:30:57  alani
 * added reconstruction of certReq to setCertReqOrig if
 * the type is pkcs10
 *
 * Revision 1.47  2002/02/24 06:42:17  urut
 * extracting of public key from pkcs#10 changed (only for v1!)
 *
 * Revision 1.46  2002/02/23 22:59:42  alani
 * added extraction of subject to setCertReq
 *
 * Revision 1.44  2002/02/22 12:38:55  urut
 * modifications to process SUCCESS response to v2 request
 *
 * Revision 1.43  2002/02/22 10:35:31  urut
 * first v2 working release (pkcs#10 only)
 * major mods,in particular:
 * - md5 hash calculation for proxyIdentifier
 * - requestPayload, Attributes is tagged object
 *
 * Revision 1.42  2002/02/22 07:34:37  alani
 * refactored setCaCert and fixed get & set StateData
 *
 * Revision 1.41  2002/02/22 03:18:07  urut
 * adjusted scep library to openscep 0.4.0:
 * -proxyIdentification Envelope Removed
 * -calculation of hash for requestPayload changed
 * -messageTyp in v2 is different
 *
 * Revision 1.40  2002/02/22 01:31:38  alani
 * added new methods setSubject(String) and setScepSubject(String)
 *
 * Revision 1.39  2002/02/16 01:14:45  urut
 * replaced all read-only accessed files with URL's file:///......
 * removed option caCert , only caCertURL is available now
 *
 * Revision 1.38  2002/02/16 00:40:20  urut
 * awfull lot of fixes,among them:
 * - computation of transactionID for v2
 * - spkac processing in cmd line client
 * - new filetyp unknown for debug files
 * - subject in cmd client
 * - lots of fixes for v2 protocoll handling
 *
 * Revision 1.37  2002/02/15 13:21:12  urut
 * first complete implementation of scep v2
 *
 * Revision 1.36  2002/02/15 10:12:20  urut
 * changed endEntityxxxx -> scepxxxx (because of v2 scep)
 *
 * Revision 1.35  2002/02/15 06:11:09  urut
 * merge with alani
 * lots of mods for incorporating full support for HTTP communication with SCEPD; Httpd client is now capable of reading multipart-mime (sort of....??) messages
 *
 * Revision 1.34  2002/02/14 16:39:40  alani
 * added a few methods & attributes needed by the http proxy
 *
 * Revision 1.33  2002/02/13 12:55:22  alani
 * some changes
 *
 * Revision 1.32  2002/02/12 19:33:23  urut
 * replaced all println statements
 *
 * Revision 1.31  2002/02/12 14:01:17  urut
 * cmd client improved
 *
 * Revision 1.30  2002/02/07 17:08:50  urut
 * ability to send scep request over http added
 *
 * Revision 1.29  2002/02/07 13:16:34  urut
 * finished http CA cert get
 *
 * Revision 1.28  2002/02/07 12:40:56  urut
 * moved util-stuff to Utils.java
 * CA get implemented
 *
 * Revision 1.27  2002/01/31 14:07:32  urut
 * adjustments for new SignedData (that takes certificates as Sequences)
 *
 * Revision 1.26  2002/01/31 11:29:45  urut
 * new cmd option "-certificate"
 *
 * Revision 1.25  2002/01/24 00:11:31  urut
 * new flag -response added
 *
 * Revision 1.24  2002/01/23 19:10:24  urut
 * fixed computation of MD5 fingerprint
 *
 * Revision 1.23  2002/01/17 19:44:00  urut
 * added attribute certReqOrig, which contains the original PKCS#10 request (because CertificationRequest is buggy)
 *
 * Revision 1.22  2002/01/17 11:52:37  urut
 * extract publicKey from pkcs#10 request if available, if not load it from file (new function, called from CmdClient )
 *
 * Revision 1.21  2002/01/15 21:20:11  urut
 * Modified EvelopedData: is now embedded in ia ContentInfo structure (required by PKCS7)
 * Modified transactionID (set to fix id for testing because it needs to be printable)
 * added missing attributes to signerInfo
 * added messageTyp=19 for PKCSReq
 * ScepClient uses now DERObject as attribute type to store envelope.
 *
 * Revision 1.20  2002/01/10 20:22:01  urut
 * bug fixing
 *
 * Revision 1.19  2002/01/10 16:17:40  urut
 * fixed some bugs
 *
 * Revision 1.18  2002/01/10 15:54:23  urut
 * load pkcs#10
 *
 * Revision 1.17  2002/01/10 15:09:38  alani
 * new attribute distinguishedName in ScepClient, scepReq now generates a transactionId
 *
 * Revision 1.15  2002/01/10 14:43:44  urut
 * merge
 *
 * Revision 1.14  2002/01/10 14:37:26  urut
 * merge
 *
 * Revision 1.13  2002/01/10 14:19:43  alani
 * improved certRequest
 *
 * Revision 1.10  2002/01/08 21:00:29  urut
 * added javadoc header with cvs tags to every file
 * Revision 1.9  2002/01/08 11:41:19  urut
 * added tags
 * </pre>
 *
 * @since $Date: 2002/02/28 01:22:16 $
 * @author  
 * @version $Revision: 1.54 $
 */

public class ScepClient implements Serializable {

	static final String COMMAND_PACKAGE = "ch.othello.openscep.internal.commands.";

	private X509CertificateStructure signedEndEntityCertificate = null;  //signed CA returned by a SUCCESS response

	private X509CertificateStructure caCert = null;
	private String caUrl = null;
	private TBSCertificateStructure tbsCACert =  null;

	private DERObject pkcsCertReqEnvelope = null;

	private DERPrintableString transactionId = null;
	private DERPrintableString challenge = null ;

	private RSAPublicKeyStructure caPubKey = null;

	private RSAPrivateKeyStructure scepPrivateKey = null;
	private RSAPublicKeyStructure scepPublicKey = null;

	private CertificationRequest certReq = null;
	private DERObject proxyCertReq = null;
	private DERObject certReqOrig = null;
        
        private X509Name subject = null;           //subject of the end entity, extracted from PKCS#10 or manually set
        
	private String url = null;
	private String request = null ; //where to store request

        /* V 2---------------*/
        private X509Name subjectScep = null;       //subject of the proxy, just used in case of version 2
        private DERObject requestPayload = null;
        private DEROctetString proxyIdentification = null;
        
        /* V 2---------------*/
        
	private SecureRandom secureRandom;

	private ContentInfo pkiMessage;

	private DERObject responseFile;

        private String communityString;
        
	SignedData pkcsCertReqSigned;

	DERInteger serial;

        public static int TYPE_PKCS10  =0;
        public static int TYPE_SPKAC   =1;
        
        public static int VERSION1     =1;
        public static int VERSION2     =2;
        
	private int requestType = TYPE_PKCS10;
	private int scepVersion = VERSION1;

        private URL certificate = null;      //answer/result is written to this URL if provided
        
    private StateData lnkStateData;

	public void setCaCert(X509CertificateStructure caCert) {
		try {
			this.caCert = caCert;
			this.tbsCACert = caCert.getTBSCertificate();
			this.caPubKey = new RSAPublicKeyStructure((DERConstructedSequence) tbsCACert.getSubjectPublicKeyInfo().getPublicKey());
			System.out.println("hello");
			if ( this.caPubKey == null )
				System.out.println("SturmBanner");
		} catch(IOException e) {
			e.printStackTrace();
		}
	}

	public X509CertificateStructure getCaCert() {
		return caCert;
	}

	/**
	 * Gets the certReqOrig
	 * @return Returns a CertificationRequest
	 */
	public DERObject getCertReqOrig() {
		return certReqOrig;
	}

	/**
	 * Sets the certReqOrig.
	 * @param certReqOrig The certReqOri to set
	 */
	public void setCertReqOrig(DERObject certReqOrig) {
                Log.debug("setCertReqOrig DERObject");
                
		this.certReqOrig = certReqOrig;
		
		if ( getRequestType() == TYPE_PKCS10 ) {
                        Log.debug("this is PKCS#10, set certReq");
                        
			setCertReq( new CertificationRequest((DERConstructedSequence)certReqOrig) );
		}
		

	}
	
	public void setCertReqOrig(String certReqOrig) {
		Log.debug("setCertReqOrig String");
                try {
			setCertReqOrig(PEMReader.decodeBase64DER( certReqOrig ));
		} catch ( IOException ex ) {
			ex.printStackTrace();
		}
	}

	/**
	 * Gets the certReq.
	 * @return Returns a CertificationRequest
	 */
	public CertificationRequest getCertReq() {
		return certReq;
	}

	/**
	 * Sets the certReq.
	 * @param certReq The certReq to set
	 */
	public void setCertReq(CertificationRequest certReq) {
		Log.debug("setCertReq CertificationRequest");
                
                this.certReq = certReq;

		try {
                //extract subject if PKCS#10
                if ( requestType==ScepClient.TYPE_PKCS10 ) {
                    subject = certReq.getCertificationRequestInfo().getSubject();
                    
                    if( scepVersion == VERSION1 ) {
                    	setSubjectScep( certReq.getCertificationRequestInfo().getSubject()); //issuer and subject is the same in V1
                    	Log.info("extracted subject from PKCS#10");
                    }
                } 
                

                if ( scepVersion == ScepClient.VERSION1  ) {
                    
                    scepPublicKey =new RSAPublicKeyStructure( (DERConstructedSequence) certReq.getCertificationRequestInfo().getSubjectPublicKeyInfo().getPublicKey());
                    Log.info("extracted publicKey from PKCS#10");
                }
                
                if (  scepVersion == ScepClient.VERSION2  && this.requestType == ScepClient.TYPE_PKCS10 ) {
                    subject = certReq.getCertificationRequestInfo().getSubject();
                    
                    
                    //TODO:this shouldn't be here,but needs to stay till bug in openScep is out
                    setSubjectScep( certReq.getCertificationRequestInfo().getSubject()); //issuer and subject is the same in V1
                    Log.info("extracted subject from PKCS#10");
                } 
  
		} catch ( IOException ex ) {
			ex.printStackTrace();
		}
		
	}
	
	public void setCertReq(String certReq) {
                Log.debug("setCertReq String");
                
                try {
                    DERObject certReqDecoded = (DERConstructedSequence)PEMReader.decodeBase64DER( certReq);
                    setCertReqOrig ( certReqDecoded );
                    setCertReq( new CertificationRequest( (DERConstructedSequence)certReqDecoded));
		} catch ( IOException ex ) {
			ex.printStackTrace();
		}
	}

	/**
	 * Gets the pkcsCertReqEnvelope.
	 * @return Returns a DERObject
	 */
	public DERObject getCertReqEnvelope() {
		return pkcsCertReqEnvelope;
	}

	/**
	 * Sets the pkcsCertReqEnvelope.
	 * @param DERObject The pkcsCertReqEnvelope to set
	 */
	public void setCertReqEnvelope(DERObject envelope) {
		this.pkcsCertReqEnvelope = envelope;
	}

	public void setTransactionId(DERPrintableString transId) {
		transactionId = transId;
	}

	public DERPrintableString getTransactionId() {
		return transactionId;
	}

	/**
	 * Gets the caPubKey.
	 * @return Returns a RSAPublicKeyStructure
	 */
	public RSAPublicKeyStructure getCaPubKey() {
		return caPubKey;
	}

	/**
	 * Sets the caPubKey.
	 * @param caPubKey The caPubKey to set
	 */
	public void setCaPubKey(RSAPublicKeyStructure caPubKey) {
		this.caPubKey = caPubKey;
	}

	/**
	 * Gets the scepPrivateKey.
	 * @return Returns a RSAPrivateKeyStructure
	 */
	public RSAPrivateKeyStructure getScepPrivateKey() {

		return scepPrivateKey;
	}

	/**
	 * Sets the scepPrivateKey.
	 * @param scepPrivateKey The scepPrivateKey to set
	 */
	public void setScepPrivateKey(RSAPrivateKeyStructure scepPrivateKey) {
		this.scepPrivateKey = scepPrivateKey;
	}

	/**
	 * Gets the scepPublicKey.
	 * @return Returns a RSAPublicKeyStructure
	 */
	public RSAPublicKeyStructure getScepPublicKey() {
		return scepPublicKey;
	}

	/**
	 * Sets the scepPublicKey.
	 * @param scepPublicKey The scepPublicKey to set
	 */
	public void setScepPublicKey(RSAPublicKeyStructure scepPublicKey) {
		this.scepPublicKey = scepPublicKey;
	}

	/**
	 * Gets the challenge.
	 * @return Returns a DERPrintableString
	 */
	public DERPrintableString getChallenge() {
		return challenge;
	}

	/**
	 * Sets the challenge.
	 * @param challenge The challenge to set
	 */
	public void setChallenge(DERPrintableString challenge) {
		this.challenge = challenge;
	}

	/**
	 * Sets the challenge.
	 * @param challenge The challenge to set
	 */
	public void setChallenge(String challenge) {
		this.challenge = new DERPrintableString(challenge);
	}

	/**
	 * Gets the url.
	 * @return Returns a String
	 */
	public String getUrl() {
		return url;
	}

	/**
	 * Sets the url.
	 * @param url The url to set
	 */
	public void setUrl(String url) {
		this.url = url;
	}

	/**
	 * Gets the secureRandom.
	 * @return Returns a SecureRandom
	 */
	public SecureRandom getSecureRandom() {
		return secureRandom;
	}

	/**
	 * Sets the secureRandom.
	 * @param secureRandom The secureRandom to set
	 */
	public void setSecureRandom(SecureRandom secureRandom) {
		this.secureRandom = secureRandom;
	}

	public void loadResponseFile(String filename)
		throws FileNotFoundException, IOException {

		responseFile = ((DERConstructedSequence) PEMReader.readFile(filename));

		Log.info("responseFile decoded as DER string");
		Log.debug(ASN1Dump.dumpAsString(responseFile.getDERObject()));

	}
	
	
	public boolean loadCA(String url ) throws IOException {
		caUrl = url;
	
		return loadCA( new URL( url ) );

	}

	/**
	 *
	 *Load CA Cert from URL
	 */
	private boolean loadCA(URL url) throws IOException {
		// Open the CA Cert
        InputStream iStream = Utils.connectToURL(url);
		DERObject obj = PEMReader.readStream(iStream);
		 
		setCaCert( new X509CertificateStructure((DERConstructedSequence) obj ));		

		Log.info("CA cert decoded as DER string");
		Log.debug(ASN1Dump.dumpAsString(caPubKey.getDERObject()));
                
		return true;
	}

	/**
	 *
	 *send DERObject to scep server (http req)
	 *
	 */
	public int post(DERObject der, String operation) throws IOException {
		Log.info("post to " + url + " with operation " + operation);

		int code = 999;

                Log.info("--- begin pkiMessage that will be sent to SCEP server");
		Log.info(ASN1Dump.dumpAsString(der ));
                Log.info("--- end pkiMessage");
                
		//base64
		try {

			ScepHTTPGet scepHttp =
				new ScepHTTPGet(
					PEMWriter.encodeByteArray(Utils.convertDERObjectToByteArray(der)),
					operation,
					url);

			code = scepHttp.getResponseCode();
			Log.info(" return code from scep httpd server : " + code);

                        if ( code == HttpURLConnection.HTTP_OK ) {
                            //read
                            byte[] resp = scepHttp.getContent();                            
                            PEMWriter.generateRequestFile( "/tmp/resp.pem", resp, PEMWriter.FILETYPE_CERTIFICATE);
                            responseFile = ((DERConstructedSequence) Utils.convertByteArrayToDERObject(resp) );
                            Log.info("response received, decoded and set");
                        }
                        
			scepHttp.disconnect();
		} catch (MalformedURLException e1) {
			Log.error("invalid URL specified");
			code = 999;
		} catch (Exception e) {
			Log.error("problem connecting to SCEP server");
			code = 999;
		}

		return (code);

	}

	/**
	 * Load CA Certificate from File
	 *
	 */
	/*public void loadCA(String filename) throws FileNotFoundException, IOException {
		// Open the CA Cert

		caCert =
			new X509CertificateStructure(
				(DERConstructedSequence) PEMReader.readFile(filename));
		// Decode the CA Certificate
		tbsCACert = caCert.getTBSCertificate();

		caPubKey =
			new RSAPublicKeyStructure(
				(DERConstructedSequence) tbsCACert.getSubjectPublicKeyInfo().getPublicKey());

		Log.info("CA cert decoded as DER string");
		Log.debug(ASN1Dump.dumpAsString(caPubKey.getDERObject()));
	}*/

	public void loadPrivateKey(URL url)
		throws FileNotFoundException, IOException {

                Log.debug("read private key from " + url);                
                InputStream iStream = Utils.connectToURL(url);
                
		scepPrivateKey = new RSAPrivateKeyStructure( (DERConstructedSequence) PEMReader.readStream(iStream));

		Log.info("scep PrivateKey decoded as DER string");
		Log.debug(ASN1Dump.dumpAsString(scepPrivateKey.getDERObject()));
	}

	public void loadPublicKey(URL url)
		throws FileNotFoundException, IOException {

                Log.debug("read publickey from " + url);
                InputStream iStream = Utils.connectToURL(url);
                
                SubjectPublicKeyInfo subjectPublicKeyInfo = new SubjectPublicKeyInfo( (DERConstructedSequence) PEMReader.readStream (iStream) );
		scepPublicKey = new RSAPublicKeyStructure( (DERConstructedSequence)subjectPublicKeyInfo.getPublicKey() );

		Log.info("scep PublicKey decoded as DER string");
		Log.debug(ASN1Dump.dumpAsString(scepPublicKey.getDERObject()));
	}

	public void loadRequest(URL url)
		throws FileNotFoundException, IOException {

		Log.debug("set certReq");

                InputStream iStream = Utils.connectToURL(url);

                if ( this.getRequestType() == ScepClient.TYPE_PKCS10 ) {
                    certReq = new CertificationRequest((DERConstructedSequence) PEMReader.readStream(iStream) );
            	    
                }
                

                InputStream iStream2 = Utils.connectToURL(url);
		certReqOrig = (((DERConstructedSequence) PEMReader.readStream(iStream2)));
                Log.debug("certReq == " + ASN1Dump.dumpAsString(certReqOrig.getDERObject()));

		Log.debug("certReqOrig " + ASN1Dump.dumpAsString(certReqOrig));

                //extract subject if PKCS#10
                if ( scepVersion == ScepClient.VERSION1  ) {                    
                    scepPublicKey =new RSAPublicKeyStructure( (DERConstructedSequence) certReq.getCertificationRequestInfo().getSubjectPublicKeyInfo().getPublicKey());
                    Log.info("extracted publicKey from PKCS#10");
                }
                
                if (  scepVersion == ScepClient.VERSION2  && this.requestType == ScepClient.TYPE_PKCS10 ) {
                    subject = certReq.getCertificationRequestInfo().getSubject();
                    
                    
                    //TODO: this should not be necessary but there is a bug in OpenSCEP that requires the subject of the proxy to be the same then the end entity
                    //setSubjectScep( certReq.getCertificationRequestInfo().getSubject()); //issuer and subject is the same in V1
                    Log.info("extracted subject from PKCS#10");
                } 
                
                //set subject to string from cmd line
                if ( scepVersion == ScepClient.VERSION2 && requestType == ScepClient.TYPE_SPKAC ) {
                    Log.debug("SPKAC & V2, subject needs to be set manually");
                }
                
		Log.info("certRequest decoded as DER string");
	}

	public static void debugOut(String str) {
		Log.debug("debug -- " + str);
	}

	public boolean Execute(String command)
		throws IOException, ScepError, ScepPending {

		try {

			Class cmdClass = Class.forName(COMMAND_PACKAGE + command);
			ScepCommand cmd = (ScepCommand) cmdClass.newInstance();

			if (cmd.Assert(this)) {
				cmd.Execute(this);
			} else {
				return false;
			}

			return true;

		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			return false;
		} catch (IllegalAccessException e) {
			e.printStackTrace();
			return false;
		} catch (InstantiationException e) {
			e.printStackTrace();
			return false;
		}

	}

	/** Getter for property responseFile.
	 * @return Value of property responseFile.
	 */
	public org.bouncycastle.asn1.DERObject getResponseFile() {
		return responseFile;
	}

	/** Setter for property responseFile.
	 * @param responseFile New value of property responseFile.
	 */
	public void setResponseFile(org.bouncycastle.asn1.DERObject responseFile) {
		this.responseFile = responseFile;
	}

	/** Getter for property signedEndEntityCertificate.
	 * @return Value of property signedEndEntityCertificate.
	 */
	public org
		.bouncycastle
		.asn1
		.x509
		.X509CertificateStructure getSignedEndEntityCertificate() {
		return signedEndEntityCertificate;
	}
	
	public String getBase64SignedEndEntityCertificate() {
		
		byte[] cert = null;
		try {
			cert = PEMWriter.encodeByteArray( 
					Utils.convertDERObjectToByteArray( signedEndEntityCertificate.getDERObject() ) );
		} catch(IOException e) {
		}
		
		return new String( cert );
		
	}

	/** Setter for property signedEndEntityCertificate.
	 * @param signedEndEntityCertificate New value of property signedEndEntityCertificate.
	 */
	public void setSignedEndEntityCertificate(
		org
			.bouncycastle
			.asn1
			.x509
			.X509CertificateStructure signedEndEntityCertificate) {
		this.signedEndEntityCertificate = signedEndEntityCertificate;
	}

	public void storeCertificate(String filename) throws Exception {
		PEMWriter.generateFile(
			filename,
			signedEndEntityCertificate.getDERObject(),
			PEMWriter.FILETYPE_CERTIFICATE);
	}

	/** Getter for property pkiMessage.
	 * @return Value of property pkiMessage.
	 */
	public org.bouncycastle.asn1.pkcs.ContentInfo getPkiMessage() {
		return pkiMessage;
	}

	public void setPkiMessage(org.bouncycastle.asn1.pkcs.ContentInfo pkiMessage) {
		this.pkiMessage = pkiMessage;
	}

	/** Getter for property request.
	 * @return Value of property request.
	 */
	public java.lang.String getRequest() {
		return request;
	}

	public void setRequest(String request) {
		this.request = request;
	}

	public ScepClient() {

		secureRandom = new SecureRandom();
		/*
		 * This following call to setSeed() makes the
		 * initialisation of the SecureRandom object
		 * _very_ fast, but not secure AT ALL.
		 * remove after development time
		 *
		 **/
		secureRandom.setSeed("www.easc.org".getBytes());

                //TODO:read from properties
                setSubjectScep( new X509Name("CN=openscepproxy"));
                
	}

	/**
	 * Gets the requestType.
	 * @return Returns a int
	 */
	public int getRequestType() {
		return requestType;
	}

	/**
	 * Sets the requestType.
	 * @param requestType The requestType to set
	 */
	public void setRequestType(int requestType) {
		this.requestType = requestType;
	}

	/**
	 * Gets the scepVersion.
	 * @return Returns a int
	 */
	public int getScepVersion() {
		return scepVersion;
	}

	/**
	 * Sets the scepVersion.
	 * @param scepVersion The scepVersion to set
	 */
	public void setScepVersion(int scepVersion) {
		this.scepVersion = scepVersion;
	}

        /** Getter for property communityString.
         * @return Value of property communityString.
         */
        public java.lang.String getCommunityString() {
            return communityString;
        }
        
        public void setCommunityString( String communityString) {
            this.communityString = communityString;
        }
        
        /** Getter for property requestPayload.
         * @return Value of property requestPayload.
         */
        public org.bouncycastle.asn1.DERObject getRequestPayload() {
            return requestPayload;
        }

        public void setRequestPayload( DERObject requestPayload) {
            this.requestPayload = requestPayload;
        }
        
        /** Getter for property proxyIdentification.
         * @return Value of property proxyIdentification.
         */
        public org.bouncycastle.asn1.DEROctetString getProxyIdentification() {
            return proxyIdentification;
        }
        
        /** Setter for property proxyIdentification.
         * @param proxyIdentificationSequence New value of property proxyIdentification.
         */
        public void setProxyIdentification (org.bouncycastle.asn1.DEROctetString proxyIdentification) {
            this.proxyIdentification = proxyIdentification;
        }
        
        /** Getter for property subjectScep.
         * @return Value of property subjectScep.
         */
        public org.bouncycastle.asn1.x509.X509Name getSubjectScep() {
            return subjectScep;
        }
        
        public void setSubjectScep( X509Name subjectScep) {
            Log.debug("setSubjectScep called - X509Name");
            this.subjectScep=subjectScep;
        }
       
       	public void setSubjectScep( String subject ) {
            Log.debug("setSubjectScep called -String");
            this.subjectScep=new X509Name( subject );
       	}
        
        /** Getter for property subject.
         * @return Value of property subject.
         */
        public org.bouncycastle.asn1.x509.X509Name getSubject() {
            return subject;
        }
        
        public void setSubject(String subject ) {
            Log.debug("setSubjectScep called -String");
            this.subject=new X509Name( subject );
        }
        
        public void setSubject(X509Name subject) {
            Log.debug("setSubjectScep called - X509Name");
            this.subject=subject;
        }
        
        public StateData getStateData() throws ScepError {
        	
        	if ( requestType==TYPE_PKCS10 && ( certReq==null || certReqOrig == null )) {
        		throw new ScepError( "Internal Error : Certificate Request structure is null" );
        	}
        	

        	return new StateData( scepVersion, requestType, caUrl,
        						  certReqOrig, subject, subjectScep, scepPublicKey,
        						  scepPrivateKey, communityString, url );
        						  
        					
        }
        
        public void setStateData( StateData data ) {
        
        	try {
                Log.debug("setStateData called");
		Log.debug("StateData - Version " + data.getVersion() );
                Log.debug("StateData - requesttype " + data.getRequestType());
                
        	setRequestType(data.getRequestType());
        	setScepVersion(data.getVersion());
        	setCommunityString(data.getCommunityString());
        	setScepPublicKey(data.getPubKey());
                loadCA( data.getCaCertUrl() );
        	setScepPrivateKey(data.getPrivKey());
                setCertReqOrig(data.getCertReqOrig());
        	setUrl(data.getUrl());

        	
        	setSubject(data.getSubject());
                
                if ( this.getRequestType() == ScepClient.TYPE_SPKAC)
                    setSubjectScep(data.getScepSubject());      //just needs to be here till openscep bug is out
        	} catch ( IOException ex ) {
            	ex.printStackTrace();
            }

        }       	
        
        /** Getter for property certificate.
         * @return Value of property certificate.
         */
        public java.net.URL getCertificate() {
            return certificate;
        }        

        public void setCertificate(URL certificate) {
            this.certificate=certificate;
        }
}