package ch.othello.openscep.internal.commands;

import java.io.*;
import java.text.*;
import java.util.*;
import java.net.*;

import org.bouncycastle.asn1.DERConstructedSequence;
import org.bouncycastle.asn1.DERConstructedSet;
import org.bouncycastle.asn1.DERInteger;
import org.bouncycastle.asn1.DERObjectIdentifier;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.DEROutputStream;
import org.bouncycastle.asn1.DERPrintableString;

import org.bouncycastle.asn1.pkcs.*;

import org.bouncycastle.asn1.util.ASN1Dump;
import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
import org.bouncycastle.asn1.x509.X509CertificateStructure;
import org.bouncycastle.crypto.AsymmetricBlockCipher;
import org.bouncycastle.crypto.BlockCipher;
import org.bouncycastle.crypto.BufferedBlockCipher;
import org.bouncycastle.crypto.CryptoException;
import org.bouncycastle.crypto.Digest;
import org.bouncycastle.crypto.KeyGenerationParameters;
import org.bouncycastle.crypto.digests.MD5Digest;
import org.bouncycastle.crypto.engines.DESEngine;
import org.bouncycastle.crypto.engines.RSAEngine;
import org.bouncycastle.crypto.generators.DESKeyGenerator;
import org.bouncycastle.crypto.modes.CBCBlockCipher;
import org.bouncycastle.crypto.paddings.PaddedBufferedBlockCipher;
import org.bouncycastle.crypto.params.DESParameters;
import org.bouncycastle.crypto.params.RSAKeyParameters;
import org.bouncycastle.crypto.encodings.*;
import org.bouncycastle.util.encoders.*;

import ch.othello.openscep.*;
import ch.othello.openscep.internal.*;

import org.bouncycastle.asn1.x509.*;
import org.bouncycastle.asn1.*;
/**
 * 
 * <pre>
 * $Id: certRequest.java,v 1.36 2002/02/27 06:53:19 urut Exp $
 * $Log: certRequest.java,v $
 * Revision 1.36  2002/02/27 06:53:19  urut
 * modified setting of subject in Signed structure
 *
 * Revision 1.35  2002/02/27 04:21:27  urut
 * first working spkac version
 *
 * Revision 1.34  2002/02/25 04:23:39  urut
 * first working version for MS browsers
 * there is still a problem in here as subjectSCEP needs to be set to the subject
 * of the actual PKCS#10 request. seems to be a bug in OpenScep.
 *
 * Revision 1.33  2002/02/24 05:18:56  urut
 * write pkcs#10 to file
 *
 * Revision 1.32  2002/02/23 23:00:27  alani
 * improved error handling
 *
 * Revision 1.31  2002/02/23 00:01:35  alani
 * new asserts
 *
 * Revision 1.30  2002/02/22 13:40:00  urut
 * v2 with spkac (sending)
 *
 * Revision 1.29  2002/02/22 10:35:31  urut
 * first v2 working release (pkcs#10 only)
 * major mods,in particular:
 * - md5 hash calculation for proxyIdentifier
 * - requestPayload, Attributes is tagged object
 *
 * Revision 1.26  2002/02/22 03:18:07  urut
 * adjusted scep library to openscep 0.4.0:
 * -proxyIdentification Envelope Removed
 * -calculation of hash for requestPayload changed
 * -messageTyp in v2 is different
 *
 * Revision 1.25  2002/02/16 01:14:45  urut
 * replaced all read-only accessed files with URL's file:///......
 * removed option caCert , only caCertURL is available now
 *
 * Revision 1.24  2002/02/16 00:40:20  urut
 * awfull lot of fixes,among them:
 * - computation of transactionID for v2
 * - spkac processing in cmd line client
 * - new filetyp unknown for debug files
 * - subject in cmd client
 * - lots of fixes for v2 protocoll handling
 *
 * Revision 1.23  2002/02/15 13:21:12  urut
 * first complete implementation of scep v2
 *
 * Revision 1.22  2002/02/15 10:12:20  urut
 * changed endEntityxxxx -> scepxxxx (because of v2 scep)
 *
 * Revision 1.21  2002/02/12 19:33:23  urut
 * replaced all println statements
 *
 * Revision 1.20  2002/02/12 14:01:17  urut
 * cmd client improved
 *
 * Revision 1.19  2002/02/07 17:08:50  urut
 * ability to send scep request over http added
 *
 * Revision 1.18  2002/01/31 14:07:32  urut
 * adjustments for new SignedData (that takes certificates as Sequences)
 *
 * Revision 1.17  2002/01/24 22:18:11  urut
 * lots of modifications; read response (CertRep), RSA decrypt, DES-cbc decrypt,
 * bouncy fix for reading SET/SEQUENCE with length=0
 *
 * Revision 1.16  2002/01/23 13:31:26  urut
 * transactionID modified. ID is MD5 fingerprint of SubjectPublicKeyInfo of PKCS#10 request, converted to (uppercase) HEX
 *
 * Revision 1.15  2002/01/17 19:48:17  urut
 * various bug fixes:
 * -padding for DES encryption
 * -OID for cbc-des
 * -padding for contentkey-encryption
 * -message digest generation for SignerInfo
 * -ContentInfo for SignedData
 * -use original DERObject instead of CertificationRequest for PKCS#10
 * -issuer=subject in self-signed certificate
 * -lots of small fixes
 *
 * Revision 1.14  2002/01/15 21:20:11  urut
 * Modified EvelopedData: is now embedded in a ContentInfo structure (required by PKCS7)
 * Modified transactionID (set to fix id for testing because it needs to be printable)
 * added missing attributes to signerInfo
 * added messageTyp=19 for PKCSReq
 * ScepClient uses now DERObject as attribute type to store envelope.
 *
 * Revision 1.13  2002/01/15 13:58:43  urut
 * Fixed AuthenticatedAttributes (each value of an attribute has to be in a SET)
 * Fixed subjectPublicKeyInfo - includes now public key of PKCS#10 request
 * Fixed issuerAndSerialNumber - was missing due to bug in SignerInfo
 * introduced unstructuredName as additional attribute
 *
 * Revision 1.12  2002/01/14 18:31:44  urut
 * Envelope is now a OctetString
 *
 * Revision 1.11  2002/01/10 20:22:01  urut
 * bug fixing
 *
 * Revision 1.10  2002/01/10 16:17:40  urut
 * fixed some bugs
 *
 * Revision 1.9  2002/01/10 15:51:13  alani
 * fixed PKCS10Generate, removed generation of PKCS10 in certReq
 *
 * Revision 1.8  2002/01/10 15:33:43  urut
 * merge
 *
 * Revision 1.7  2002/01/10 15:31:25  urut
 * pkcs10 generate
 *
 * Revision 1.5  2002/01/10 14:49:51  alani
 * fixed errors from merge
 *
 * </pre>
 *
 * @since $Date: 2002/02/27 06:53:19 $
 * @version $Revision: 1.36 $
 * @author alani
 */

public class certRequest implements ScepCommand {

	ScepClient client;

	/* calculate MD5 fingerprint of SubjectPublicKey and convert it to Hex*/
	private void generateTransactionId() throws IOException {

		ByteArrayOutputStream bOut = new ByteArrayOutputStream();

		DEROutputStream dOut = new DEROutputStream(bOut);

		if (client.getScepVersion() == ScepClient.VERSION1) {
			Log.info("computing transactionID for V1");
			dOut.writeObject(
				client
					.getCertReq()
					.getCertificationRequestInfo()
					.getSubjectPublicKeyInfo()
					.getDERObject());
		} else if (client.getScepVersion() == ScepClient.VERSION2) {
			Log.info("computing transactionID for V2");

			//V2 transaction ID contains both: SCEP proxy public key and request (orig)
			dOut.writeObject(client.getScepPublicKey().getDERObject());
			dOut.writeObject(client.getCertReqOrig());
		} else {
			Log.error("unknown version");
		}

		//dOut.writeObject( client.getCertReqOrig() );

		//endEntityPublicKey

		Digest digest = new MD5Digest();
		digest.update(bOut.toByteArray(), 0, bOut.size());
		byte[] hash = new byte[digest.getDigestSize()];
		digest.doFinal(hash, 0);

		//make HEX string out of it
		HexTranslator translator = new HexTranslator();
		byte[] hex = new byte[2 * hash.length];
		translator.encode(hash, 0, hash.length, hex, 0);

		ByteArrayOutputStream bOutHex = new ByteArrayOutputStream();
		bOutHex.write(hex);

		Log.debug("fingerprint is: " + bOutHex.toString().toUpperCase());

		//RSAMD5Digest.printOutByteArray( hex );

		client.setTransactionId(
			new DERPrintableString(bOutHex.toString().toUpperCase()));

	}

	private SignedData createSigned() throws IOException, ScepError {

		DERConstructedSet algorithms = new DERConstructedSet();

		algorithms.addObject(new AlgorithmIdentifier(PKCSObjectIdentifiers.md5));

		ContentInfo contentInfo =
			new ContentInfo(
				PKCSObjectIdentifiers.data,
				new DEROctetString(client.getCertReqEnvelope()));

		DERConstructedSequence certificates = new DERConstructedSequence();

		//create self signed certificate
		V3TBSCertificateGenerator tbs = new V3TBSCertificateGenerator();

		/*
		 *    subjectPublicKeyInfo {
		 *       algorithm        pkcs-1 1
		 *       subjectPublicKey BITSTRING
		 *    }
		 */
		SubjectPublicKeyInfo subjectPublicKeyInfo =
			new SubjectPublicKeyInfo(
				new AlgorithmIdentifier(
					new DERObjectIdentifier(PKCSObjectIdentifiers.pkcs_1 + ".1")),
				client.getScepPublicKey());
		tbs.setSubjectPublicKeyInfo(subjectPublicKeyInfo);

		SimpleDateFormat dateF = new SimpleDateFormat("yyMMddHHmmss");
		SimpleTimeZone tz = new SimpleTimeZone(0, "Z");
		dateF.setTimeZone(tz);

		tbs.setEndDate(
			new DERUTCTime(
				dateF.format(new Date(System.currentTimeMillis() + 4000000)) + "Z"));
		;

		tbs.setStartDate(new DERUTCTime(dateF.format(new Date()) + "Z"));
		;

		if (client.getSubjectScep() == null) {
			Log.error("SubjectScep is NULL");
			throw new ScepError("subjectScep is NULL");
		}

		tbs.setSubject(client.getSubjectScep());

		//TODO: use real serial number instead of 4711
		tbs.setSerialNumber(new DERInteger(4711)); //=transactionId
		tbs.setSignature(
			new AlgorithmIdentifier(
				new DERObjectIdentifier(PKCSObjectIdentifiers.pkcs_1 + ".4")));


		tbs.setIssuer(client.getSubjectScep());
		//*tbs.setIssuer ( client.getCertReq().getCertificationRequestInfo().getSubject() );

		DERConstructedSequence selfSignedCert = new DERConstructedSequence();

		TBSCertificateStructure tbsToSign = tbs.generateTBSCertificate();

		selfSignedCert.addObject(tbsToSign);
		selfSignedCert.addObject(
			new AlgorithmIdentifier(
				new DERObjectIdentifier(PKCSObjectIdentifiers.pkcs_1 + ".4")));

		//signature of Certificate is BitString

		Log.debug("sign self signed certificate before");

		byte[] signature =
			RSAMD5Digest.signMD5RSA(tbsToSign, client.getScepPrivateKey());

		Log.debug("now trying to decode again.....");
		//try to decrypt for testing
		//RSAMD5Digest.decryptMD5( signature, client.getScepPublicKey() );

		selfSignedCert.addObject(new DERBitString(signature));

		Log.debug("sign self signed certificate after");

		certificates.addObject(selfSignedCert);

		//compute messagedigest (authenticated attribute)
		DEROctetString messageDigest =
			new DEROctetString(RSAMD5Digest.digestMD5(client.getCertReqEnvelope()));

		DERConstructedSet authAttribs = new DERConstructedSet();
		DERConstructedSequence att = new DERConstructedSequence();
		att.addObject(new DERObjectIdentifier(PKCSObjectIdentifiers.pkcs_9 + ".3"));
		//pkcs-9 3

		/*  structure of authenticated attributes
		 
		    Sequence ::= {
		       Set ::= {
		         pkcs-7 1
		       }
		    }
		 */

		DERConstructedSet contentType = new DERConstructedSet();
		contentType.addObject(PKCSObjectIdentifiers.data); // pkcs-7 1
		att.addObject(contentType);
		authAttribs.addObject(att);

		//messageDigest
		att = new DERConstructedSequence();
		att.addObject(new DERObjectIdentifier(PKCSObjectIdentifiers.pkcs_9 + ".4"));
		// pkcs-9 4
		DERConstructedSet messageDigestAttribute = new DERConstructedSet();
		messageDigestAttribute.addObject(messageDigest);
		att.addObject(messageDigestAttribute);
		authAttribs.addObject(att);

		//transactionID
		att = new DERConstructedSequence();
		att.addObject(SCEPObjectIdentifiers.id_transId); // 2.16.840.1.113733.1.9.7
		DERConstructedSet transactionId = new DERConstructedSet();
		transactionId.addObject(client.getTransactionId());
		att.addObject(transactionId);
		authAttribs.addObject(att);

		//messageType
		att = new DERConstructedSequence();
		att.addObject(SCEPObjectIdentifiers.id_messageType);
		DERConstructedSet messageType = new DERConstructedSet();

		//messageTyp is scep version specific, v1 is 19 fix (PKCSReq), v2 is 17,18 or 19    
		DERPrintableString messageTypeString = null;
		if (client.getScepVersion() == ScepClient.VERSION1) {
			messageTypeString = new DERPrintableString("19");
		} else if (client.getScepVersion() == ScepClient.VERSION2) {
			messageTypeString = new DERPrintableString("18");
		} else {
			Log.error("invalid version,can't continue");
		}
		messageType.addObject(messageTypeString);
		// PrintableString, 19=PKCSReq (according SCEP proposal), 18=request from proxy, 17=request with requestPayload structure
		att.addObject(messageType);
		authAttribs.addObject(att);

		//senderNonce 
		att = new DERConstructedSequence();
		att.addObject(SCEPObjectIdentifiers.id_senderNonce);
		DERConstructedSet senderNonceAttribute = new DERConstructedSet();
		byte[] senderNonce = new byte[16];
		client.getSecureRandom().nextBytes(senderNonce);
		senderNonceAttribute.addObject(new DEROctetString(senderNonce));
		att.addObject(senderNonceAttribute);
		authAttribs.addObject(att);

		//additional attributes in case of V2
		if (client.getScepVersion() == ScepClient.VERSION2) {
			Log.info("this is version 2, add additional attributes");

			//request type
			/*att = new DERConstructedSequence();
			att.addObject(SCEPObjectIdentifiers.id_requestType);
			DERConstructedSet requestType = new DERConstructedSet();
			
			if ( client.getRequestType() == ScepClient.TYPE_PKCS10 ) { 
			    requestType.addObject( new DERPrintableString("1") );   // PrintableString
			} else if ( client.getRequestType() == ScepClient.TYPE_SPKAC ) {
			    requestType.addObject( new DERPrintableString("2") );   // PrintableString                        
			} else {
			    Log.error("invalid request type,not supported");
			}
			
			att.addObject( requestType );
			authAttribs.addObject(att);
			*/

			//proxy identification
			att = new DERConstructedSequence();
			att.addObject(SCEPObjectIdentifiers.id_proxyIdentification);
			DERConstructedSet proxyIdentification = new DERConstructedSet();

			proxyIdentification.addObject(client.getProxyIdentification()); // Octet String

			att.addObject(proxyIdentification);
			authAttribs.addObject(att);

		}

		// now we should digest & encrypt the authenticated attribs ( with caCert )

		DEROctetString encryptedDigest =
			new DEROctetString(
				RSAMD5Digest.digestMD5RSA(authAttribs, client.getScepPrivateKey()));

		//TODO: add real issuer name
		SignerInfo info = new SignerInfo(
		/*client.getCertReq().getCertificationRequestInfo().getSubject(),*/
		client.getSubjectScep(), /*issuer, =subject in V1*/
		/*client.getIssuerSerial(),*/
			new DERInteger(4711), //,
		new AlgorithmIdentifier(PKCSObjectIdentifiers.md5), // 113549.2.5 = md5
	authAttribs,
		new AlgorithmIdentifier(PKCSObjectIdentifiers.rsaEncryption),
		encryptedDigest,
		null);

		DERConstructedSet signerInfos = new DERConstructedSet();
		signerInfos.addObject(info);

		SignedData signed =
			new SignedData(algorithms, contentInfo, certificates, null, signerInfos);

		//write out various debug output for debugging
		try {
			PEMWriter.generateRequestFile(
				"certreqenv.pem",
				client.getCertReqEnvelope(),
				PEMWriter.FILETYPE_CERTIFICATE_REQUEST);
			PEMWriter.generateRequestFile(
				"signerInfos.pem",
				signerInfos.getDERObject(),
				PEMWriter.FILETYPE_CERTIFICATE_REQUEST);
			PEMWriter.generateRequestFile(
				"contentinfo.pem",
				contentInfo.getDERObject(),
				PEMWriter.FILETYPE_CERTIFICATE_REQUEST);
			PEMWriter.generateRequestFile(
				"certificates.pem",
				certificates.getDERObject(),
				PEMWriter.FILETYPE_CERTIFICATE_REQUEST);
			PEMWriter.generateRequestFile(
				"signed.pem",
				signed.getDERObject(),
				PEMWriter.FILETYPE_CERTIFICATE_REQUEST);
		} catch (Exception e) {
		}

		return signed;
	}

	private DERObject createEnvelope() {

		X509CertificateStructure caCert = client.getCaCert();

		DERConstructedSequence reInfo = new DERConstructedSequence();
		reInfo.addObject(new DERInteger(0)); // the version for recipientInfo

		DERConstructedSequence issuerAndSerialNumber = new DERConstructedSequence();
		issuerAndSerialNumber.addObject(caCert.getIssuer());
		issuerAndSerialNumber.addObject(caCert.getSerialNumber());

		reInfo.addObject(issuerAndSerialNumber);

		//reInfo.addObject(caCert.getIssuer()); // These two are X509Names OK ??
		//reInfo.addObject(caCert.getSerialNumber());

		Log.debug(" Serial Number is " + ASN1Dump.dumpAsString(caCert.getSerialNumber()));

		DERConstructedSequence algIdentifiers = new DERConstructedSequence();
		algIdentifiers.addObject(PKCSObjectIdentifiers.rsaEncryption);
		reInfo.addObject(algIdentifiers);

		// First Generate a Random Key for DESEncryption DESede
		byte[] desKey = CryptUtils.generateRandomDESKey(client.getSecureRandom());
		byte[] encryptedKey =
			CryptUtils.RSAencryptWithPadding(desKey, client.getCaPubKey());

		/*DESKeyGenerator desgen = new DESKeyGenerator();
		KeyGenerationParameters kgp =
		new KeyGenerationParameters(client.getSecureRandom(), DESParameters.DES_KEY_LENGTH * 8);
		desgen.init(kgp);
		byte[] desKey = desgen.generateKey();
		*/

		//then make a RSA-Encryption with the desKey

		/*
		AsymmetricBlockCipher rsaEngWithoutPadding = new RSAEngine();
		
		AsymmetricBlockCipher rsaEng = new PKCS1Encoding ( rsaEngWithoutPadding );
		
		
		RSAKeyParameters pubParameters =
		new RSAKeyParameters(false, client.getCaPubKey().getModulus(),
						  client.getCaPubKey().getPublicExponent());
		rsaEng.init(true, pubParameters);
		byte[] encryptedKey;
		
		try {
		encryptedKey = rsaEng.processBlock(desKey, 0, desKey.length);
		reInfo.addObject(new DEROctetString(encryptedKey));
		} catch (Exception e) {
		e.printStackTrace();
		}*/

		reInfo.addObject(new DEROctetString(encryptedKey));

		// make a ByteArray from the CertReq
		DEROctetString content = null;
		try {
			ByteArrayOutputStream bOut = new ByteArrayOutputStream();
			DEROutputStream dOut = new DEROutputStream(bOut);

                        //write debug
                        try {
        		PEMWriter.generateRequestFile("/tmp/pkcs10Debug.pem", client.getCertReqOrig(), PEMWriter.FILETYPE_UNKNOWN);                
                        } catch(Exception eee){}
                        
			if (client.getScepVersion() == ScepClient.VERSION1) {

                            Log.debug(
					"got pkcs#10 request: " + ASN1Dump.dumpAsString(client.getCertReqOrig()));
				dOut.writeObject(client.getCertReqOrig());
			} else {
				//V2, requestPayload
				Log.debug("got payload: " + ASN1Dump.dumpAsString(client.getRequestPayload()));
				dOut.writeObject(client.getRequestPayload());
			}

			byte[] input = bOut.toByteArray();
			byte[] out = CryptUtils.DESencryptWithPadding(input, desKey);
			content = new DEROctetString(out);
		} catch (IOException e) {
			e.printStackTrace();
		}
		//catch (CryptoException e) {
		// throw an Exception here
		//	e.printStackTrace();
		//}

		Log.debug("encrypted pkcs (V1) or requestPayload (V2) is:");
		Log.debug(ASN1Dump.dumpAsString(content));

		DERConstructedSet reSet = new DERConstructedSet();
		reSet.addObject(reInfo);

			EnvelopedData envelope = new EnvelopedData(reSet, //recipientInfo
		PKCSObjectIdentifiers.data, //pkcs-7 1
	new AlgorithmIdentifier(SCEPObjectIdentifiers.des_cbc), content);
		// encryptedContent

		ContentInfo contentInfoEnveloped =
			new ContentInfo(PKCSObjectIdentifiers.envelopedData, envelope.getDERObject());

		//DERConstructedSequence contentInfo = new DERConstructedSequence();
		//contentInfo.addObject( contentInfoEnveloped );

		return contentInfoEnveloped.getDERObject();

	}

	public boolean Assert(ScepClient theClient) throws IOException {
		client = theClient;

		if ( client.getRequestType() != client.TYPE_PKCS10 && client.getScepVersion() == client.VERSION1 ) {
			Log.error("SCEP Version 1 only accepts PKCS10 Requests");
			return false;
		}

		//either -response or -url needs to be specified
		if (client.getUrl() == null) {
			Log.error("Invalid parameters, URL needs to be specified");
			return false;
		}

		if (client.getCertReqOrig() == null) {
			Log.error("Invalid parameters, CertReqOrig needs to be specified");
			return false;
		}

		if (client.getCaCert() == null) {
			Log.error("CA cert needs to be specified");
			return false;
		}

		if (client.getCommunityString() == null) {
			client.setCommunityString(new String());
		}

		if (client.getRequestType() == client.TYPE_SPKAC
			&& client.getSubject() == null ) {
			Log.error("Subject needs to be specified");
			return false;
		}


		if (client.getTransactionId() == null) {
			generateTransactionId();
		}

		return true;
	}

	public ContentInfo createPKIMessage(SignedData data) {

		ContentInfo contentInfo =
			new ContentInfo(
				new DERObjectIdentifier(PKCSObjectIdentifiers.pkcs_7 + ".2"),
				data.getDERObject());

		return (contentInfo);
	}

	//TODO:add ASN.1 spec as soon as spec is final
	//only V2
	public DEROctetString generateProxyIdentification() {
		//DERConstructedSequence seq = new DERConstructedSequence();

		Log.debug("used communityString is " + client.getCommunityString());

		LinkedList list = new LinkedList();
		list.add(client.getRequestPayload()); //DEREncodable
		list.add(/*new DERPrintableString(*/
		client.getCommunityString() /*)*/
		); //string

		DEROctetString digest = new DEROctetString(RSAMD5Digest.digestMD5(list));

		//seq.addObject(  digest  );

		//TODO : probably remove communityString (?)
		//seq.addObject( new DERIA5String( client.getCommunityString() ) );

		Log.debug("generateProxyIdentification successfull");

		/*try{
		   PEMWriter.generateRequestFile("/tmp/proxyIdentification.pem", seq, PEMWriter.FILETYPE_UNKNOWN);                
		   Log.debug("debug written to /tmp/proxyIdentification.pem");
		} catch ( Exception e) {e.printStackTrace();}
		*/

		return (digest);
	}

	//TODO:add ASN.1 spec as soon as spec is final
	//only V2
	//this encrypts proxyIdentificationSequence,generates a random DES key, encrypts DES key with RSA
	/*        public DERObject generateProxyIdenificationData(DERObject proxyIdentificationSequence) {
	            
			X509CertificateStructure caCert = client.getCaCert();
	
	                DERConstructedSequence reInfo = new DERConstructedSequence();       //recipientInfo
			reInfo.addObject(new DERInteger(0)); // the version for recipientInfo
	                
	                
	                DERConstructedSequence issuerAndSerialNumber = new DERConstructedSequence();
	                issuerAndSerialNumber.addObject( caCert.getIssuer()  );
	                issuerAndSerialNumber.addObject( caCert.getSerialNumber()  );
	                
	                reInfo.addObject( issuerAndSerialNumber );
	                
	                Log.debug( "CA Serial Number is " + ASN1Dump.dumpAsString(caCert.getSerialNumber()  ) );
	                
	                DERConstructedSequence algIdentifiers = new DERConstructedSequence();
	                algIdentifiers.addObject ( PKCSObjectIdentifiers.rsaEncryption );
			reInfo.addObject( algIdentifiers );
			
			// First Generate a Random Key for DESEncryption DESede
	                byte[] desKey = CryptUtils.generateRandomDESKey( client.getSecureRandom() );
	                
			//then make a RSA-Encryption with the desKey
	                byte[] encryptedKey = CryptUtils.RSAencryptWithPadding( desKey, client.getCaPubKey() );
	                
	                reInfo.addObject(new DEROctetString( encryptedKey ));
	                
			// make a ByteArray from the CertReq
			DEROctetString content = null;
			try {
				ByteArrayOutputStream bOut = new ByteArrayOutputStream();
				DEROutputStream dOut = new DEROutputStream(bOut);
	                        Log.debug("got pkcs#10 request: " + ASN1Dump.dumpAsString( proxyIdentificationSequence ));
				dOut.writeObject( proxyIdentificationSequence );
	                        			
	                        byte[] input = bOut.toByteArray();                        
	                        byte[] out = CryptUtils.DESencryptWithPadding( input, desKey);
				content = new DEROctetString(out);
			} catch (IOException e) {
				e.printStackTrace();
			} 
	                //catch (CryptoException e) {
				// throw an Exception here
			//	e.printStackTrace();
			//}
	                
	                Log.debug("encrypted proxyIdentificationSequence is:");
	                Log.debug(ASN1Dump.dumpAsString(content));                
	                
			DERConstructedSet reSet = new DERConstructedSet();
			reSet.addObject(reInfo);
	
			EnvelopedData envelope =
				new EnvelopedData(
					reSet,                              //recipientInfo
					PKCSObjectIdentifiers.data,         //pkcs-7 1
					new AlgorithmIdentifier( SCEPObjectIdentifiers.des_cbc ),
					content );  // encryptedContent
	
	                ContentInfo contentInfoEnveloped = new ContentInfo(PKCSObjectIdentifiers.envelopedData, envelope.getDERObject() );
	                
	                //DERConstructedSequence contentInfo = new DERConstructedSequence();
	                //contentInfo.addObject( contentInfoEnveloped );
	               
	                Log.debug("end proxyidentificagtiondata");
	            try{
	               PEMWriter.generateRequestFile("/tmp/proxyIdentificationData.pem", contentInfoEnveloped.getDERObject(), PEMWriter.FILETYPE_UNKNOWN);                
	               Log.debug("debug written to /tmp/proxyIdentificationData.pem");
	            } catch ( Exception e) {e.printStackTrace();}
	                
	                return contentInfoEnveloped.getDERObject();
	        }
	  */

	//only V2
	public DERObject generateRequestPayload() {
		DERConstructedSequence seq = new DERConstructedSequence();

		seq.addObject(new DERInteger(client.getRequestType()));
		//version, 0=pkcs, 1=spkac
		seq.addObject(new DERBitString(client.getCertReqOrig()));

		/*  structure of attributes
		     
		        Sequence ::= {
		           Set ::= {
		             pkcs-7 1
		           }
		        }
		*/

		DERConstructedSet attributeSet = new DERConstructedSet();

		DERTaggedObject tagged = null;

		if (client.getRequestType() == ScepClient.TYPE_PKCS10) {
			Log.debug("Request type PKCS#10 , set empty attribute set in requestPayload");

			tagged = new DERTaggedObject(false, 1, attributeSet);

		} else if (client.getRequestType() == ScepClient.TYPE_SPKAC) {

			Log.info("Request type SPKAC, set supplied attributes");

			DERConstructedSequence att = null;

			DERConstructedSequence x509Att =
				(DERConstructedSequence) client.getSubject().getDERObject();

			//read all attributes
			for (int i = 0; i < x509Att.getSize(); i++) {

				DERConstructedSet derSet = (DERSet) x509Att.getObjectAt(i);
				DERConstructedSequence derSeq = (DERConstructedSequence) derSet.getObjectAt(0);

				att = new DERConstructedSequence();
				att.addObject((DERObjectIdentifier) derSeq.getObjectAt(0));
				DERConstructedSet oneAttribute = new DERConstructedSet();
				oneAttribute.addObject((DERObject) derSeq.getObjectAt(1));
				att.addObject(oneAttribute);
				attributeSet.addObject(att);
			}

			tagged = new DERTaggedObject(false, 1, attributeSet);

			//add additional attribute subject
			//attributeSeq.addObject( OID );

			//DERConstructedSet attribute = new DERConstructedSet();
			//attribute.addObject( VALUE );

			//attributeSeq.
			//set.addObject( client.getSubject() );
		}

		seq.addObject(tagged);

		try {
			PEMWriter.generateRequestFile(
				"/tmp/requestPayload.pem",
				seq,
				PEMWriter.FILETYPE_UNKNOWN);
			Log.debug("debug written to /tmp/requestPayload.pem");
		} catch (Exception e) {
			e.printStackTrace();
		}

		return (seq);
	}

	/*
	 * @see ScepCommand#Execute(ScepClient)
	 */
	public boolean Execute(ScepClient theClient)
		throws IOException, ScepError, ScepPending {

		client = theClient;

                
                //just tests....
                //client.setRequestType( ScepClient.TYPE_PKCS10);
                //client.setScepVersion(ScepClient.VERSION2);
		//client.loadPublicKey( new URL("file:///tmp/rsapublickey.pem") );
		//client.loadPrivateKey( new URL("file:///tmp/rsaprivatekey.pem") );
                
          	//client.loadCA( new URL("http://localhost/cacert.pem") );
                
                //client.loadRequest(new URL("file:///tmp/ms.pem") );
                
                
		if (client.getScepVersion() == ScepClient.VERSION2) {

			client.setRequestPayload(generateRequestPayload());
			Log.info("Payload generated");

			DEROctetString proxyIdentification = generateProxyIdentification();
			client.setProxyIdentification(proxyIdentification);
			Log.info("ProxyIdentification generated");

			//client.setProxyIdentificationData( generateProxyIdenificationData( proxyIdentificationSequence ) );
			//Log.info("ProxyIdentificationData generated");
		}

		client.setCertReqEnvelope(createEnvelope());

		Log.info("--- begin pkcsCertReqEnvelope");
		Log.info(ASN1Dump.dumpAsString(client.getCertReqEnvelope()));
		Log.info("--- end pkcsCertReqEnvelope");

		SignedData pkcsCertReqSigned = createSigned();

		Log.info("--- begin pkcsCertReqSigned");
		Log.info(ASN1Dump.dumpAsString(pkcsCertReqSigned.getDERObject()));
		Log.info("--- end pkcsCertReqSigned");

		ContentInfo pkiMessage = createPKIMessage(pkcsCertReqSigned);

		client.setPkiMessage(pkiMessage);

		Log.info("--- begin pkiMessage");
		Log.info(ASN1Dump.dumpAsString(pkiMessage.getDERObject()));
		Log.info("--- end pkiMessage");

		//store request in a file
		if (client.getRequest() != null) {
			try {
				//PEMWriter.generateRequestFile( "pkcs10.pem", client.getCertReqOrig().getDERObject(), PEMWriter.FILETYPE_CERTIFICATE_REQUEST);
				PEMWriter.generateRequestFile(
					client.getRequest(),
					pkiMessage.getDERObject(),
					PEMWriter.FILETYPE_CERTIFICATE_REQUEST);

				//PEMWriter.generateRequestFile( "signed.pem", pkcsCertReqSigned.getDERObject(), PEMWriter.FILETYPE_CERTIFICATE_REQUEST);

				//PEMWriter.generateRequestFile("envelope.pem", client.getCertReqEnvelope(), PEMWriter.FILETYPE_CERTIFICATE_REQUEST);

			} catch (Exception e) {
				Log.error("couldn't write file");
			}
		}

		if (client.getUrl() != null) {
			if (client.post(client.getPkiMessage().getDERObject(), "PKIOperation")
				== HttpURLConnection.HTTP_OK) {
				Log.info("request successfully sent to SCEP server, got response");

				Log.info("try to decode reply");
				client.Execute("certRep");

			} else {
				Log.info("couldn't send request to SCEP server (no HTTP 200!)");
				throw new ScepError(
					"couldn't send request to SCEP server (" + client.getUrl() + ")");
			}
		}

		return true;

	}

}
